from PIL import Image
from tqdm import tqdm
from typing import List
from abcli.logger import logger, crash_report


NAME = "abcli.plugins.gif"


def generate_animated_gif(
    list_of_images: List[str],
    output_filename: str,
    frame_duration: int = 150,
):
    if not list_of_images:
        return True

    logger.info(
        "{}.generate_animated_gif({} frames) -> {} @ {:.2f}ms".format(
            NAME,
            len(list_of_images),
            output_filename,
            frame_duration,
        )
    )

    max_width = 0
    max_height = 0
    frames = []
    for filename in tqdm(list_of_images):
        image = Image.open(filename)

        frames.append(image)

        width, height = image.size
        max_width = max(max_width, width)
        max_height = max(max_height, height)

    padded_frames = []
    for image in frames:
        width, height = image.size
        left_pad = (max_width - width) // 2
        top_pad = (max_height - height) // 2
        padded_image = Image.new(
            "RGB",
            (max_width, max_height),
            (255, 255, 255),
        )
        padded_image.paste(image, (left_pad, top_pad))
        padded_frames.append(padded_image)

    try:
        padded_frames[0].save(
            output_filename,
            save_all=True,
            append_images=padded_frames[1:],
            duration=frame_duration,
            loop=0,  # 0 means infinite loop
        )
    except Exception:
        crash_report("generate_animated_gif")
        return False

    logger.info(
        "generate_animated_gif: {}x{}x{} -> {}".format(
            len(list_of_images),
            height,
            width,
            output_filename,
        )
    )

    return True



from typing import Dict, List
import fnmatch
from functools import reduce
import os
import shutil
from abcli import env
from abcli import string
from .load import load_text
from .save import save_text
from . import NAME
from abcli.logger import logger, crash_report


def absolute(filename, reference_path=None):
    """return absolute path to filename.

    Args:
        filename (str): filename.
        reference_path (str, optional): reference path. Defaults to None.

    Returns:
        str: filename.
    """
    from abcli.path import absolute as path_absolute

    return os.path.join(
        path_absolute(
            path(filename),
            os.getcwd() if reference_path is None else reference_path,
        ),
        name_and_extension(filename),
    )


def add_postfix(filename, postfix):
    """add postfix to filename.

    Args:
        filename (str): filename.
        postfix (str): postfix.

    Returns:
        str: filename.
    """
    filename, extension = os.path.splitext(filename)
    return f"{filename}-{postfix}{extension}"


def add_prefix(filename, prefix):
    """add prefix to name of filename.

    Args:
        filename (str): filename.
        prefix (str): prefix.

    Returns:
        str: filename
    """
    pathname, filename = os.path.split(filename)
    return os.path.join(pathname, f"{prefix}-{filename}")


def auxiliary(nickname, extension, add_timestamp=True):
    """generate auxiliary filename.

    Args:
        nickname (str): nickname
        extension (str): extension
        add_timestamp (bool, optional): add timestamp. Defaults to True.

    Returns:
        str: auxiliary filename.
    """
    from abcli.file.save import prepare_for_saving

    filename = os.path.join(
        env.abcli_object_path,
        "auxiliary",
        "-".join(
            [nickname]
            + (
                [
                    string.pretty_date(
                        as_filename=True,
                        squeeze=True,
                        unique=True,
                    )
                ]
                if add_timestamp
                else []
            )
        )
        + f".{extension}",
    )

    prepare_for_saving(filename)

    return filename


def build_from_template(
    template_filename: str,
    segments: Dict[str, List[str]],
    filename: str,
) -> bool:
    success, template = load_text(template_filename)
    if not success:
        return success

    content: List[str] = []
    for line in template:
        line_completed: List[str] = [line]

        for template_line in segments:
            if template_line in line:
                line_completed = segments[template_line]
                break

        content += line_completed

    return save_text(filename, content)


def copy(
    source,
    destination,
    log=True,
    overwrite=True,
):
    """copy source to destination.

    Args:
        source (str): source filename.
        destination (str): destination filename.
        log (bool, optional): log. Defaults to True.
        overwrite (bool, optional): log. Defaults to True.

    Returns:
        bool: success
    """
    from abcli.file.save import prepare_for_saving

    if not overwrite and exist(destination):
        if log:
            logger.info(f"✅ {destination}")
        return True

    if not prepare_for_saving(destination):
        return False

    try:
        # https://stackoverflow.com/a/8858026
        # better choice: copy2
        shutil.copyfile(source, destination)
    except:
        crash_report(f"-{NAME}: copy({source},{destination}): failed.")
        return False

    if log:
        logger.info(f"{NAME}: {source} -> {destination}")

    return True


def create(filename, content=[]):
    """create filename.

    Args:
        filename (str): filename.
        content (list, optional): content. Defaults to [].

    Returns:
        bool: success
    """
    from abcli.file.save import save_text

    return save_text(filename, content)


def delete(filename):
    """delete filename.

    Args:
        filename (str): filename.

    Returns:
        bool: success.
    """
    if not os.path.isfile(filename):
        return True

    try:
        os.remove(filename)

        return True
    except:
        crash_report(f"-{NAME}: delete({filename}): failed.")
        return False


def download(
    url,
    filename,
    log=True,
    overwrite=True,
):
    """download url as filename.

    Args:
        url (str): url.
        filename (str): filename.
        log (bool, optional): log. Defaults to True.
        overwrite (bool, optional): overwrite. Defaults to True.

    Returns:
        bool: success
    """
    if not overwrite and exist(filename):
        return True

    try:
        import urllib3

        # https://stackoverflow.com/a/27406501
        with urllib3.PoolManager().request(
            "GET", url, preload_content=False
        ) as response, open(filename, "wb") as fp:
            shutil.copyfileobj(response, fp)

        response.release_conn()  # not 100% sure this is required though

    except:
        crash_report(f"-{NAME}: download({url},{filename}): failed.")
        return False

    if log:
        logger.info(f"{NAME}: {url} -> {filename}")

    return True


def exist(filename):
    """does filename exist?

    Args:
        filename (str): filename.

    Returns:
        bool: does filename exist?
    """
    return os.path.isfile(filename)


def extension(filename):
    """return extension of filename.

    Args:
        filename (str): filename.

    Returns:
        str: extension.
    """

    if isinstance(filename, str):
        _, extension = os.path.splitext(filename)
        if extension != "":
            if extension[0] == ".":
                extension = extension[1:]
        return extension

    if isinstance(filename, type):
        return "py" + filename.__name__.lower()

    return "py" + filename.__class__.__name__.lower()


def list_of(template, recursive=False):
    """return list of files that match template.

    Args:
        template (str): template.
        recursive (bool, optional): recursive. Defaults to False.

    Returns:
        List[str]: list of filenames.
    """
    from abcli import path as abcli_path

    if isinstance(template, list):
        return reduce(
            lambda x, y: x + y,
            [list_of(template_, recursive) for template_ in template],
            [],
        )

    if recursive:
        return reduce(
            lambda x, y: x + y,
            [
                list_of(
                    os.path.join(pathname, name_and_extension(template)),
                    recursive,
                )
                for pathname in abcli_path.list_of(path(template))
            ],
            list_of(template),
        )

    # https://stackoverflow.com/a/40566802
    template_path = path(template)
    if template_path == "":
        template_path = abcli_path.current()

    try:
        return [
            os.path.join(template_path, filename)
            for filename in fnmatch.filter(
                os.listdir(template_path),
                name_and_extension(template),
            )
        ]
    except:
        return []


def move(source, destination):
    """move source to destination.

    Args:
        source (str): source.
        destination (str): destination.

    Returns:
        bool: success.
    """
    from abcli.file.save import prepare_for_saving

    if not prepare_for_saving(destination):
        return False

    try:
        # https://stackoverflow.com/a/8858026
        shutil.move(source, destination)
    except:
        crash_report(f"-{NAME}: move({source},{destination}): failed.")
        return False

    return True


def name(filename):
    """return name of filename.

    Args:
        filename (str): filename.

    Returns:
        str: name of filename.
    """
    _, filename = os.path.split(filename)

    return filename if "." not in filename else ".".join(filename.split(".")[:-1])


def name_and_extension(filename):
    """return name and extension of filename.

    Args:
        filename (str): filename.

    Returns:
        str: name and extension of filename
    """
    return os.path.basename(filename)


def path(filename):
    """return path of filename.

    Args:
        filename (str): filename.

    Returns:
        str: path of filename.
    """
    return os.path.split(filename)[0]


def relative(filename, reference_path=None):
    """return relative filename.

    Args:
        filename (str): filename.
        reference_path (str, optional): reference path. Defaults to None.

    Returns:
        str: relative filename.
    """

    from abcli.path import relative as path_relative

    return path_relative(
        path(filename),
        os.getcwd() if reference_path is None else reference_path,
    ) + name_and_extension(filename)


def set_extension(filename, extension_, force=True):
    """set extension for filename.

    Args:
        filename (str): filename.
        extension_ (str): extension.
        force (bool, optional): force. Defaults to True.

    Returns:
        str: filename.
    """
    if not isinstance(extension_, str):
        extension_ = extension(extension_)

    filename, extension_as_is = os.path.splitext(filename)
    if extension_as_is != "":
        extension_as_is = extension_as_is[1:]

    if not force and extension_as_is == "":
        extension_ = extension_as_is

    return f"{filename}.{extension_}"


def size(filename):
    """return size of file.

    Args:
        filename (str): filename.

    Returns:
        int: size of filename in bytes.
    """
    try:
        return os.path.getsize(filename)
    except:
        return 0



from abcli.string import random_
from abcli.plugins.markdown import generate_table


def test_generate_table():
    assert generate_table(
        [random_() for _ in range(50)],
        cols=4,
    )



import pytest
from typing import Callable
from abcli.plugins.metadata import (
    get,
    post,
    get_from_file,
    get_from_object,
    get_from_path,
    post,
    post_to_file,
    post_to_object,
    post_to_path,
    MetadataSourceType,
)
from abcli.modules.objects import unique_object, object_path, path_of
from abcli.string import random_


@pytest.mark.parametrize(
    [
        "post_source",
        "post_source_type",
    ],
    [
        [
            lambda object_name: path_of(
                filename="metadata.yaml",
                object_name=object_name,
            ),
            MetadataSourceType.FILENAME,
        ],
        [
            lambda object_name: object_name,
            MetadataSourceType.OBJECT,
        ],
        [
            lambda object_name: object_path(
                object_name=object_name,
                create=True,
            ),
            MetadataSourceType.PATH,
        ],
    ],
)
@pytest.mark.parametrize(
    [
        "get_source",
        "get_source_type",
    ],
    [
        [
            lambda object_name: path_of(
                filename="metadata.yaml",
                object_name=object_name,
            ),
            MetadataSourceType.FILENAME,
        ],
        [
            lambda object_name: object_name,
            MetadataSourceType.OBJECT,
        ],
        [
            lambda object_name: object_path(
                object_name=object_name,
                create=True,
            ),
            MetadataSourceType.PATH,
        ],
    ],
)
def test_metadata(
    post_source: Callable[[str], str],
    post_source_type: MetadataSourceType,
    get_source: Callable[[str], str],
    get_source_type: MetadataSourceType,
):
    object_name = unique_object()
    key = random_()
    value = random_()

    assert post(
        key=key,
        value=value,
        source=post_source(object_name),
        source_type=post_source_type,
    )

    returned_value = get(
        key=key,
        source=get_source(object_name),
        source_type=get_source_type,
    )

    assert value == returned_value


def test_metadata_dict():
    object_name = unique_object()
    key = random_()
    value = {random_(): random_() for _ in range(10)}

    assert post(
        key=key,
        value=value,
        source=object_name,
        source_type=MetadataSourceType.OBJECT,
    )

    assert (
        get(
            key=key,
            source=object_name,
            source_type=MetadataSourceType.OBJECT,
        )
        == value
    )

    assert sorted(
        get(
            key=key,
            source=object_name,
            source_type=MetadataSourceType.OBJECT,
            dict_keys=True,
        )
    ) == sorted(value.keys())

    assert sorted(
        get(
            key=key,
            source=object_name,
            source_type=MetadataSourceType.OBJECT,
            dict_values=True,
        )
    ) == sorted(value.values())


def test_metadata_file():
    object_name = unique_object()
    key = random_()
    value = random_()

    filename = path_of(
        filename="metadata.yaml",
        object_name=object_name,
    )

    assert post_to_file(filename, key, value)

    assert get_from_file(filename, key) == value


def test_metadata_object():
    object_name = unique_object()
    key = random_()
    value = random_()

    assert post_to_object(object_name, key, value)

    assert get_from_object(object_name, key) == value


def test_metadata_path():
    object_name = unique_object()
    key = random_()
    value = random_()

    path = object_path(object_name=object_name, create=True)

    assert post_to_path(path, key, value)

    assert get_from_path(path, key) == value



import pytest
import numpy as np
from abcli import file
from abcli.modules import objects
from abcli.plugins.graphics.signature import add_signature
from abcli import env
from abcli.plugins.testing import download_object


# https://www.randomtextgenerator.com/
dummy_text = "This is some dummy text. This is some dummy text. This is some dummy text. This is some dummy text. This is some dummy text. This is some dummy text. This is some dummy text. This is some dummy text. This is some dummy text. This is some dummy text."


@pytest.mark.parametrize(
    ["object_name"],
    [
        [env.VANWATCH_TEST_OBJECT],
    ],
)
def test_add_signature(object_name):
    assert download_object(object_name)

    success, image = file.load_image(
        objects.path_of(
            "Victoria41East.jpg",
            object_name,
        )
    )
    assert success

    output_image = add_signature(
        image,
        header=[dummy_text],
        footer=[dummy_text, dummy_text],
    )

    assert isinstance(output_image, np.ndarray)



import argparse
from . import *
from abcli.logger import logger
from blueness.argparse.generic import sys_exit


parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    default="get",
    help="clone|create|get|list|search|set",
)
parser.add_argument(
    "--count",
    type=int,
    default=-1,
)
parser.add_argument(
    "--delim",
    type=str,
    default=", ",
)
parser.add_argument(
    "--filename",
    type=str,
)
parser.add_argument(
    "--item_name",
    default="object",
    type=str,
)
parser.add_argument(
    "--log",
    default=1,
    type=int,
    help="0|1",
)
parser.add_argument(
    "--object_1",
    type=str,
)
parser.add_argument(
    "--object_2",
    type=str,
)
parser.add_argument(
    "--relation",
    default="",
    type=str,
)
parser.add_argument(
    "--return_list",
    type=int,
    default=0,
)
args = parser.parse_args()

delim = " " if args.delim == "space" else args.delim

success = False
if args.task == "clone":
    success = clone(args.object_1, args.object_2)
elif args.task == "create":
    success = create()
elif args.task == "get":
    relation = get(args.object_1, args.object_2)
    if args.log:
        logger.info(
            f"{args.object_1} -{f'{relation}->' if relation else 'X-'} {args.object_2}"
        )
    else:
        print(relation)
    success = True
elif args.task == "list":
    for thing in sorted(
        ["{} : {}".format(this, that) for this, that in inverse_of.items() if this]
    ):
        print(thing)
    success = True
elif args.task == "search":
    output = search(args.object_1, args.relation, args.count)

    if args.relation:
        if args.log:
            logger.info(f"{len(output):,} {args.item_name}(s): {delim.join(output)}")
        else:
            print(delim.join(output))
    else:
        print(output)

    success = file.save_json(args.filename, output) if args.filename else True
elif args.task == "set":
    success = set_(args.object_1, args.object_2, args.relation)
else:
    success = None

sys_exit(logger, NAME, args.task, success, log=args.log)



import argparse
from . import *

list_of_tasks = "after|before|pretty_date|random"


parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help=list_of_tasks,
)
parser.add_argument(
    "--include_time",
    type=int,
    help="0|1",
    default=1,
)
parser.add_argument(
    "--length",
    type=int,
    default=16,
)
parser.add_argument(
    "--string",
    type=str,
)
parser.add_argument(
    "--substring",
    type=str,
)
parser.add_argument(
    "--unique",
    type=int,
    help="0|1",
    default=0,
)
args = parser.parse_args()

success = args.task in list_of_tasks.split("|")
if args.task == "after":
    print(after(args.string, args.substring))
elif args.task == "before":
    print(before(args.string, args.substring))
elif args.task == "pretty_date":
    print(
        pretty_date(
            as_filename=True,
            include_time=args.include_time,
            unique=args.unique,
        )
    )
elif args.task == "random":
    print(random_(args.length))
else:
    print(f"-{NAME}: {args.task}: command not found")

if not success:
    print(f"-{NAME}: {args.task}: failed")



from typing import List
from blueness import module
from abcli import NAME as MY_NAME
from abcli import fullname
from abcli import file
from abcli.plugins import markdown
from abcli.logger import logger

MY_NAME = module.name(__file__, MY_NAME)


def build(
    items: List[str],
    template_filename: str,
    filename: str,
    NAME: str,
    VERSION: str,
    REPO_NAME: str,
    cols: int = 3,
):
    logger.info(
        "{}.build: {} -{}-{}-@-{}-#{}-> {}".format(
            MY_NAME,
            template_filename,
            NAME,
            VERSION,
            REPO_NAME,
            cols,
            filename,
        )
    )

    table = markdown.generate_table(items, cols=cols)

    signature = [
        "---",
        "built by [`{}`]({}), based on [`{}-{}`]({}).".format(
            fullname(),
            "https://github.com/kamangir/awesome-bash-cli",
            NAME,
            VERSION,
            f"https://github.com/kamangir/{REPO_NAME}",
        ),
    ]

    return file.build_from_template(
        template_filename,
        {
            "--table--": table,
            "--signature": signature,
        },
        filename,
    )



from typing import Any
from .enums import MetadataSourceType
import base64
import copy
from abcli import file
from abcli.plugins import NAME
from abcli.logger import logger

NAME = f"{NAME}.metadata"


def post(
    key: str,
    value: Any,
    filename: str = "metadata.yaml",
    source=".",
    source_type: MetadataSourceType = MetadataSourceType.FILENAME,
    is_base64_encoded=False,
    verbose: bool = False,
) -> bool:
    if is_base64_encoded:
        value = str(base64.b64decode(value))

    filename = source_type.filename(source, filename)

    _, metadata = file.load_yaml(filename, civilized=True)

    metadata[key] = copy.deepcopy(value)

    logger.info(
        "{}.post[{}]: {}{}".format(
            NAME,
            filename,
            key,
            f"={value}" if verbose else "",
        )
    )

    return file.save_yaml(filename, metadata)


def post_to_file(
    filename: str,
    key: str,
    value: Any,
    **kwargs,
) -> bool:
    return post(
        key=key,
        value=value,
        source=filename,
        source_type=MetadataSourceType.FILENAME,
        **kwargs,
    )


def post_to_object(
    object_name: str,
    key: str,
    value: Any,
    **kwargs,
) -> bool:
    return post(
        key=key,
        value=value,
        source=object_name,
        source_type=MetadataSourceType.OBJECT,
        **kwargs,
    )


def post_to_path(
    path: str,
    key: str,
    value: Any,
    **kwargs,
) -> bool:
    return post(
        key=key,
        value=value,
        source=path,
        source_type=MetadataSourceType.PATH,
        **kwargs,
    )



import argparse
from abcli.plugins.git import version, NAME
from abcli.logger import logger
from blueness.argparse.generic import sys_exit


parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    default="",
    help="increment_version",
)
parser.add_argument(
    "--repo_path",
    type=str,
)
parser.add_argument(
    "--verbose",
    type=int,
    default="0",
    help="0|1",
)
args = parser.parse_args()

success = False
if args.task == "increment_version":
    success = version.increment(
        repo_path=args.repo_path,
        verbose=args.verbose == 1,
    )
else:
    success = None

sys_exit(logger, NAME, args.task, success)



from typing import List, Dict
import math
from abcli.logger import logger


def generate_table(items: List[str], cols=3) -> List[str]:
    if not items:
        return []

    items_dict = {index: item for index, item in enumerate(items)}

    cols = min(cols, len(items))

    row_count = int(math.ceil(len(items) / cols))

    logger.info(
        "markdown.generate_table(): {} item(s), {} row(s)".format(
            len(items),
            row_count,
        )
    )

    return [
        "| {} |".format(" | ".join(cols * [" "])),
        "| {} |".format(" | ".join(cols * ["---"])),
    ] + [
        "| {} |".format(
            " | ".join(
                [items_dict.get(cols * row_index + index, "") for index in range(cols)]
            )
        )
        for row_index in range(row_count)
    ]
